#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : basellm
# @Time         : 2023/10/13 10:48
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *

from langchain.chat_models.openai import ChatOpenAI
from langchain.adapters.openai import convert_message_to_dict
from langchain.schema.messages import BaseMessage


class BaseLLM(ChatOpenAI):

    @property
    def _llm_type(self) -> str:
        return Path(__file__).name[:-3]

    def get_num_tokens_from_messages(self, messages: List[BaseMessage]) -> int:
        """Calculate num tokens with tiktoken package.

        Official documentation: https://github.com/openai/openai-cookbook/blob/
        main/examples/How_to_format_inputs_to_ChatGPT_models.ipynb"""
        if sys.version_info[1] <= 7:
            return super().get_num_tokens_from_messages(messages)
        model, encoding = self._get_encoding_model()
        tokens_per_message = 3
        tokens_per_name = 1
        num_tokens = 0
        messages_dict = [convert_message_to_dict(m) for m in messages]
        for message in messages_dict:
            num_tokens += tokens_per_message
            for key, value in message.items():
                # Cast str(value) in case the message value is not a string
                # This occurs with function messages
                num_tokens += len(encoding.encode(str(value)))
                if key == "name":
                    num_tokens += tokens_per_name
        # every reply is primed with <im_start>assistant
        num_tokens += 3
        return num_tokens



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : openai_types
# @Time         : 2023/12/19 09:46
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai.types.chat.completion_create_params import CompletionCreateParams, CompletionCreateParamsStreaming

completion_keys = set(CompletionCreateParamsStreaming.__annotations__.keys())


class SpeechCreateRequest(BaseModel):
    input: str
    model: str = 'tts'
    voice: str = "alloy"
    response_format: Literal["mp3", "opus", "aac", "flac"] = "mp3"
    speed: float = 1


chat_completion = {
    "id": "chatcmpl-id",
    "object": "chat.completion",
    "created": 0,
    "model": "LLM",
    "choices": [
        {
            "message": {"role": "assistant", "content": ''},
            "index": 0,
            "finish_reason": "stop",
            "logprobs": None
        }
    ],
    "usage": {"completion_tokens": 0, "prompt_tokens": 0, "total_tokens": 0}

}

chat_completion_chunk = {
    "id": "chatcmpl-id",
    "object": "chat.completion.chunk",
    "created": 0,
    "model": "LLM",
    "choices": [
        {
            "delta": {"role": "assistant", "content": ''},
            "index": 0,
            "finish_reason": "stop",
            "logprobs": None
        }
    ]
}

# 通用
chat_completion = ChatCompletion.model_validate(chat_completion)
chat_completion_chunk = ChatCompletionChunk.model_validate(chat_completion_chunk)

# ONEAPI_SLOGAN
ONEAPI_SLOGAN = os.getenv("ONEAPI_SLOGAN", "\n\n[永远相信美好的事情即将发生](https://api.chatllm.vip/)")

chat_completion_slogan = chat_completion.model_copy(deep=True)
chat_completion_slogan.choices[0].message.content = ONEAPI_SLOGAN

chat_completion_chunk_slogan = chat_completion_chunk.model_copy(deep=True)
chat_completion_chunk_slogan.choices[0].delta.content = ONEAPI_SLOGAN

# ERROR
chat_completion_error = chat_completion.model_copy(deep=True)
chat_completion_chunk_error = chat_completion_chunk.model_copy(deep=True)

# PPU
chat_completion_ppu = chat_completion.model_copy(deep=True)
chat_completion_ppu.choices[0].message.content = "按次收费"
chat_completion_chunk_ppu = chat_completion_chunk.model_copy(deep=True)
chat_completion_chunk_ppu.choices[0].delta.content = "按次收费"


class ChatCompletionRequest(BaseModel):
    request: CompletionCreateParams = {}
    headers: dict = {}  # HTTPException(status.HTTP_401_UNAUTHORIZED, "Token is wrong!")


if __name__ == '__main__':
    # data = {"stream": True, "model": "gpt-3.5-turbo", "messages": [{"role": "user", "content": "你好"}]}
    # print(ChatCompletionRequest(request=data).request)
    # print(chat_completion_error)
    # print(chat_completion_slogan)

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : files
# @Time         : 2023/12/29 14:21
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  : todo: 返回链接

import shortuuid

from meutils.pipe import *
from meutils.serving.fastapi.dependencies.auth import get_bearer_token, HTTPAuthorizationCredentials
from chatllm.utils.ppu_utils import get_ppu_usage

from enum import Enum
from minio import Minio
from redis import Redis
from openai._types import FileTypes
from openai.types.file_object import FileObject
from fastapi import APIRouter, File, UploadFile, Query, Form, BackgroundTasks, Depends, HTTPException, Request, status
from fastapi.responses import Response, FileResponse

router = APIRouter()

file_map = {} or Redis(**os.getenv("REDIS_CLIENT_PARAMS", {}), decode_responses=True)  # redis_client_params


class Purpose(str, Enum):
    assistants = "assistants"
    fine_tune = "fine-tune"


minio_client = Minio(
    endpoint=os.getenv('MINIO_ENDPOINT'),
    access_key=os.getenv('MINIO_ACCESS_KEY'),
    secret_key=os.getenv('MINIO_SECRET_KEY'),
    secure=False)

OPENAI_BUCKET = os.getenv('OPENAI_BUCKET', 'openai')  # todo 根据sk区分桶


@router.get("/files/{file_id}")
async def get_files(
    file_id: str,
    auth: Optional[HTTPAuthorizationCredentials] = Depends(get_bearer_token),

):
    api_key = auth and auth.credentials or None
    if api_key is None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="认证失败")

    try:
        file_name = file_map.get(file_id)
        response = minio_client.get_object(OPENAI_BUCKET, f"{api_key}/{file_name}")
        data = response.read()

        return FileObject(
            id=file_id,
            bytes=len(data),
            created_at=int(time.time()),
            filename=file_name,
            object='file',
            purpose='assistants',
            status='uploaded',
        )
    except Exception as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=f"File error: {e}")


@router.get("/files/{file_id}/content")
async def get_files_content(
    file_id: str,
    auth: Optional[HTTPAuthorizationCredentials] = Depends(get_bearer_token),

):
    api_key = auth and auth.credentials or None
    if api_key is None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="认证失败")
    try:
        file_name = file_map.get(file_id)
        response = minio_client.get_object(OPENAI_BUCKET, f"{api_key}/{file_name}")  # 获取链接
        data = response.data

        return Response(content=data, media_type="application/octet-stream")
    except Exception as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=f"File content error: {e}")


@router.post("/files")  # 同名文件会被覆盖
async def upload_files(
    file: UploadFile = File(...),
    purpose: Purpose = Form(...),
    # return_url=Form(...)
    auth: Optional[HTTPAuthorizationCredentials] = Depends(get_bearer_token),

):
    api_key = auth and auth.credentials or None
    if api_key is None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="认证失败")

    if purpose == Purpose.fine_tune:
        # 对于微调，验证上传文件的格式
        content = await file.read()
        try:
            lines = content.decode().split("\n")
            for line in lines:
                record = json.loads(line)
                if "prompt" not in record or "completion" not in record:
                    raise ValueError("Invalid format")
        except:
            raise HTTPException(status_code=400, detail="Invalid file format for fine-tuning")
    elif purpose == Purpose.assistants:
        # 对于助手和消息，你可能需要进行不同的处理
        # todo: 直接解析到 es【后面支持队列】

        minio_client.put_object(OPENAI_BUCKET, f"{api_key}/{file.filename}", data=file.file, length=file.size)

        # fileid2filename
        file_id = f"file-{hash(file.filename)}"
        file_map[file_id] = file.filename  # shortuuid.uuid('a') "file-HASH"
        # logger.debug(file_map)

        # 按次计费
        try:
            _ = get_ppu_usage(api_key=api_key, model='ppu-file')  # 无效key
        except Exception as e:
            logger.error(e)
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="认证失败")

        return FileObject(
            # id=f"file-{shortuuid.random()}",  # 新的模型名称
            id=file_id,

            bytes=file.size,
            created_at=int(time.time()),
            filename=file.filename,
            object='file',
            purpose='assistants',
            status='uploaded',  # Deprecated
        )


if __name__ == '__main__':
    from meutils.serving.fastapi import App

    VERSION_PREFIX = '/v1'

    app = App()
    app.include_router(router, VERSION_PREFIX)
    app.run(port=9000)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : nbce
# @Time         : 2023/5/30 16:15
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :


import json
import torch
from transformers import AutoTokenizer
from transformers import LlamaForCausalLM
from transformers import TopPLogitsWarper, LogitsProcessorList
from scipy.special import logsumexp


# Renyi熵优化后的代码
# 定义Renyi熵函数
def renyi_entropy(logits, alpha=0.25):
    # 对logits进行Renyi熵优化
    k = torch.argmax(torch.exp(logits), dim=-1)
    max_logits = logits[torch.arange(logits.shape[0]), k]
    logits_uncond = logits[:, 0]
    logits_merged = (1 + alpha) * max_logits - alpha * logits_uncond
    logits = torch.where(logits_uncond > -100, logits_merged, max_logits)
    return logits


# 经过微调的LLAMA
# 下载地址：https://openbuddy.ai/
model_path = '/root/autodl-tmp/7b-trans-chat-0516-bf16'

# 加载tokenizer
tokenizer = AutoTokenizer.from_pretrained(model_path)
tokenizer.padding_side = 'left'
tokenizer.pad_token = tokenizer.unk_token

# 加载LLAMA模型
model = LlamaForCausalLM.from_pretrained(model_path, device_map='auto', torch_dtype=torch.bfloat16)
device = torch.device('cuda')

# 加载示例Context
contexts = json.load(open('contexts.json'))

# 示例问题集（一次性问多个问题，NBCE自行根据Context逐一输出答案）
question = """请仔细阅读材料，逐一回答：
- 菲律宾国家电网公司，中国占股多少？
- 领英计划裁员多少人？
- 吉利德收购Pharmasset的价格是多少？
- 丙肝神药Sovaldi在哪一年上市？
- 中亚峰会将在哪里举行？由谁主持？
- 哪个演员由于侮辱人民军队而被立案调查？
- 哪个项目宣称“能过坦克”的水上道路？
- 如果你是默沙东的CEO，你的首要任务是什么？"""

# 拼接context和question
contexts = [''] + contexts  # 添加空Context（无Context预测）
batch = ['User: %s\n\n%s\n\nAssistant:' % (context, question) for context in contexts]
print('Context长度分布：', [len(text) for text in batch])
print('Context总长度：', sum([len(text) for text in batch]))

# Top-P截断
processors = LogitsProcessorList()
processors.append(TopPLogitsWarper(0.95))


@torch.inference_mode()
def generate(max_tokens):
    """Naive Bayes-based Context Extension 演示代码
    """
    inputs = tokenizer(batch, padding='longest', return_tensors='pt').to(device)
    input_ids = inputs.input_ids
    attention_mask = inputs.attention_mask

    print('input_ids', input_ids.shape)
    past_key_values = None
    n = input_ids.shape[0]

    for i in range(max_tokens):
        # 模型输出
        outputs = model(input_ids=input_ids,
                        attention_mask=attention_mask,
                        return_dict=True,
                        use_cache=True,
                        past_key_values=past_key_values
                        )
        past_key_values = outputs.past_key_values

        # 对logits进行Renyi熵优化
        logits = outputs.logits[:, -1]
        logits = logits - logits.logsumexp(logits, dim=-1, keepdims=True)
        logits = renyi_entropy(logits)
        logits = processors(input_ids, logits)
        # # ===== 核心代码开始 =====
        # beta = 0.25
        # logits = outputs.logits[:, -1]
        # logits = logits - logits.logsumexp(dim=-1, keepdims=True)
        # logits = processors(input_ids, logits)
        # k = (logits.exp() * logits.clip(-100, 0)).sum(dim=-1)[1:].argmax() + 1
        # logits_max = logits[k]
        # logits_uncond = logits[0]
        # logits_merged = (1 + beta) * logits_max - beta * logits_uncond
        # logits = torch.where(logits_uncond > -100, logits_merged, logits_max)
        # # ===== 核心代码结束 =====

        # 构建分布，采样
        # tau = 1是标准的随机采样，tau->0则是贪心搜索
        # 简单起见，这里没有实现topk、topp截断
        tau = 0.01
        probas = torch.nn.functional.softmax(logits[None] / tau, dim=-1)
        next_tokens = torch.multinomial(probas, num_samples=1).squeeze(1)
        if next_tokens[0] == tokenizer.eos_token_id:
            break

        ret = tokenizer.batch_decode(next_tokens)
        print(ret[0], flush=True, end='')

        # prepare for next iteration
        input_ids = next_tokens.unsqueeze(-1).tile(n, 1)
        attention_mask = torch.cat([attention_mask, torch.ones(n, 1, dtype=torch.long, device=device)], dim=-1)

# # 定义Renyi熵的计算函数
# import torch
# import torch.nn.functional as F
#
# def renyi_entropy(logits, alpha):
#     probs = F.softmax(logits, dim=1)
#     p_alpha = probs ** alpha
#     sum_p_alpha = torch.sum(p_alpha, dim=1)
#     entropy = (1 / (1 - alpha)) * torch.log(sum_p_alpha)
#     return entropy
#
# # 模型输出
# outputs = model(input_ids=input_ids,
#                 attention_mask=attention_mask,
#                 return_dict=True,
#                 use_cache=True,
#                 past_key_values=past_key_values
#                )
# past_key_values = outputs.past_key_values
#
# # 提取模型输出的logits
# logits = outputs.logits[:, -1]
#
# # 计算Renyi熵
# alpha = 2.0
# logits_alpha = logits ** alpha
# sum_logits_alpha = logits_alpha.sum(dim=-1)
# logits_renyi = (1 / (1 - alpha)) * torch.log(sum_logits_alpha)
#
# # 计算调和平均数
# beta = 0.25
# logits_harmonic = (logits.exp() * logits.clip(-100, 0)).sum(dim=-1)[1:] ** (-1)
# harmonic_mean_k = torch.argmin(logits_harmonic) + 1
#
# # 计算融合后的logits
# logits_max = logits[harmonic_mean_k]
# logits_uncond = logits[0]
# logits_merged = (1 + beta) * logits_max - beta * logits_uncond
# logits = torch.where(logits_uncond > -100, logits_merged, logits_max)
#
# # 构建分布，采样
# tau = 0.01
# probas = torch.nn.functional.softmax(logits[None] / tau , dim=-1)
# next_tokens = torch.multinomial(probas, num_samples=1).squeeze(1)
# if next_tokens[0] == tokenizer.eos_token_id:
#     break
#
# ret = tokenizer.batch_decode(next_tokens)
# print(ret[0], flush=True, end='')
#
# # prepare for next iteration
# input_ids = next_tokens.unsqueeze(-1).tile(n, 1)
# attention_mask = torch.cat([attention_mask, torch.ones(n, 1, dtype=torch.long, device=device)], dim=-1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : chatbase
# @Time         : 2023/7/5 15:29
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from langchain.schema import Document

from langchain.chat_models import ChatOpenAI
from langchain.cache import InMemoryCache
from langchain.memory import ConversationBufferWindowMemory
from langchain.embeddings import OpenAIEmbeddings
from langchain.embeddings.base import Embeddings
from langchain.vectorstores import VectorStore, DocArrayInMemorySearch, Zilliz, FAISS
from langchain.callbacks import AsyncIteratorCallbackHandler

from langchain.chains.question_answering import load_qa_chain
from langchain.chains.qa_with_sources import load_qa_with_sources_chain  # 输出SOURCE废token
from langchain.chains import ConversationChain
from langchain.document_loaders import DirectoryLoader, PyMuPDFLoader


# import langchain
#
# langchain.verbose = True
# langchain.debug = True


class ChatBase(object):
    """
    ChatBase().create_index().search().run(query='1+1')
    """

    def __init__(self, model="gpt-3.5-turbo", embeddings: Embeddings = OpenAIEmbeddings(chunk_size=100), k=1,
                 temperature=0):
        self.memory = ConversationBufferWindowMemory(memory_key="chat_history", return_messages=True, k=k)
        self.memory_messages = self.memory.chat_memory.messages
        self.embeddings = embeddings  # todo: 本地向量
        self.llm = ChatOpenAI(model=model, temperature=temperature, streaming=True)
        self.chain = load_qa_chain(self.llm, chain_type="stuff")  # map_rerank 重排序
        #
        self._docs = None
        self._index = None
        self._input = None

    def create_index(self, docs: List[Document], vectorstore: VectorStore = DocArrayInMemorySearch):  # 主要耗时，缓存是否生效
        self._index = vectorstore.from_documents(docs, self.embeddings)  # 向量阶段：可以多线程走缓存？
        return self

    def search(self, query, k: int = 5, threshold: float = 0.7, **kwargs):
        docs_scores = self._index.similarity_search_with_score(query, k=k, **kwargs)
        self._docs = []
        for doc, score in docs_scores:
            if score > threshold:
                doc.metadata['score'] = score
                doc.metadata['page_content'] = doc.page_content
                self._docs.append(doc)

        self._input = {"input_documents": self._docs, "question": query}  # todo: input_func

        return self

    def run(self):
        return self.chain.run(self._input)  # 流式




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : chatbook
# @Time         : 2023/9/19 10:28
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

# 根据文章长短选用4k 16k 32k

from meutils.pipe import *

from chatllm.llmchain.applications import ChatBase


class ChatBook(ChatBase):
    pass



import base64
import hashlib
import hmac
import json
import os
from urllib.parse import urlparse
from datetime import datetime
from time import mktime
from urllib.parse import urlencode
from wsgiref.handlers import format_date_time
import requests


class EmbeddingReq(object):

    def __init__(self, appid, api_key, api_secret, embedding_url):
        self.APPID = appid
        self.APIKey = api_key
        self.APISecret = api_secret
        self.host = urlparse(embedding_url).netloc
        self.path = urlparse(embedding_url).path
        self.url = embedding_url

    # 生成url
    def create_url(self):
        # 生成RFC1123格式的时间戳
        now = datetime.now()
        date = format_date_time(mktime(now.timetuple()))

        # 拼接字符串
        signature_origin = "host: " + self.host + "\n"
        signature_origin += "date: " + date + "\n"
        signature_origin += "POST " + self.path + " HTTP/1.1"

        # 进行hmac-sha256进行加密
        signature_sha = hmac.new(self.APISecret.encode('utf-8'), signature_origin.encode('utf-8'),
                                 digestmod=hashlib.sha256).digest()

        signature_sha_base64 = base64.b64encode(signature_sha).decode(encoding='utf-8')

        authorization_origin = f'api_key="{self.APIKey}", algorithm="hmac-sha256", headers="host date request-line", signature="{signature_sha_base64}"'

        authorization = base64.b64encode(authorization_origin.encode('utf-8')).decode(encoding='utf-8')

        # 将请求的鉴权参数组合为字典
        v = {
            "authorization": authorization,
            "date": date,
            "host": self.host
        }
        # 拼接鉴权参数，生成url
        url = self.url + '?' + urlencode(v)
        # 此处打印出建立连接时候的url,参考本demo的时候可取消上方打印的注释，比对相同参数时生成的url与自己代码生成的url是否一致
        return url

    def get_Embedding(self, text):
        param_dict = {
            'header': {
                'app_id': self.APPID
            },
            'payload': {
                'text': text
            }
        }
        response = requests.post(url=self.create_url(), json=param_dict)
        result = json.loads(response.content.decode('utf-8'))
        print(result)


if __name__ == "__main__":
    # 测试时候在此处正确填写相关信息即可运行
    emb = EmbeddingReq(appid="",
                       api_key="",
                       api_secret="",
                       embedding_url=r'https://knowledge-retrieval.cn-huabei-1.xf-yun.com/v1/aiui/embedding/query'
                       )
    emb.get_Embedding('这个问题的向量是什么？')



